/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#include "Backend.h"
#include "AndroidFileResource.h"
#include "TaskItem.h"
#include "core/StdFileResource.h"
#include "core/TaskItem.h"
#include "core/TasksView.h"
#include "core/TodoMd.h"
#include "cpp-utils/debug.h"
#include <algorithm>
#include <exception>
#include <iostream>
#include <memory>
#include <ostream>
#include <qjsonarray.h>
#include <qjsonvalue.h>
#include <qlogging.h>
#include <qstandardpaths.h>
#include <qvariant.h>
#include <sstream>

Backend::Backend() : todoView(&mirai)
{
	cpputils::debug::Timer startDuration;
	std::cout << "Backend created" << std::endl;

	cpputils::debug::Timer readConfigDuration;
	QDir().mkpath(getConfigFilePath());
	QFile loadFile(getConfigFilePath());
	readConfigDuration.printTimeElapsed("Read config duration");

	if (loadFile.open(QIODevice::ReadOnly)) {
		QByteArray loadData = loadFile.readAll();
		QJsonDocument json = QJsonDocument::fromJson(loadData);
		loadFile.close();
		if (!json.isObject()) {
			qWarning() << "config.json is not a valid config file";
			exit(1);
		}
		QJsonObject jsonRootObject = json.object();

		auto jsonFilesPath = json["files"];
		if (!jsonFilesPath.isArray()) {
			qWarning() << "config.json should contains a 'files' string array";
			exit(1);
		}
		for (const QJsonValueRef &filePath : jsonFilesPath.toArray()) {
			cpputils::debug::Timer loadingFileDuration;
			auto fileResource =
				std::make_unique<mirai::AndroidFileResource>(mirai::BaseFileResourceConstructor{
					.name = "[Can't load path]", .path = filePath.toString().toStdString()
				});
			mirai.loadResource(std::move(fileResource));
			loadingFileDuration.printTimeElapsed(
				"Loading file duration of " + filePath.toString().toStdString()
			);
		}

		auto jsonTagsConfig = json["tags"];
		if (jsonTagsConfig.isObject()) {
			for (auto &jsonTagConfigKey : jsonTagsConfig.toObject().keys()) {
				tagsConfig[jsonTagConfigKey] =
					jsonTagsConfig.toObject()[jsonTagConfigKey].toObject()["color"].toString();
			}
		}
	} else {
		qWarning() << "Couldn't find existing config file";
	}

	cpputils::debug::Timer updatingViewDuration;
	todoView.update();
	updatingViewDuration.printTimeElapsed("Updating view duration");
	cpputils::debug::Timer rebuildQMLTasksListDuration;
	rebuildQMLTasksList();
	rebuildQMLTasksListDuration.printTimeElapsed("Rebuilding QML duration");
	startDuration.printTimeElapsed("Start duration");
}

void Backend::addTodoFromRawFormat(QString filePath, QString text, QString date)
{
	auto resource = std::ranges::find_if(
		mirai.getResources(),
		[&](std::unique_ptr<mirai::BaseResource> &resource) {
			auto fsResource = dynamic_cast<mirai::BaseFileResource *>(resource.get());
			if (!fsResource) {
				return false;
			}
			return fsResource->getPath() == filePath.toStdString();
		}
	);
	if (resource == mirai.getResources().end()) {
		qWarning() << "File path '" << filePath << "' doesn't exist" << Qt::endl;
		return;
	}
	resource->get()->addTask(
		mirai::TodoMdFormat::stringToTask(text.toStdString(), date.toStdString())
	);
	mirai.save();
	todoView.update();
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::addTagFilter(QString tag)
{
	todoView.addTagFilter(tag.toStdString());
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::removeTagFilter(QString tag)
{
	todoView.removeTagFilter(tag.toStdString());
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::addResourceFilter(QString fileName)
{
	todoView.addResourceFilter(fileName.toStdString());
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::removeResourceFilter(QString fileName)
{
	todoView.removeResourceFilter(fileName.toStdString());
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::removeFilters()
{
	todoView.removeFilters();
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::updateTodoFromRawFormat(int todoIndex, QString text, QString date)
{
	QMLTaskItem &taskItem = QMLTasks[todoIndex];
	taskItem.taskItem->setText(taskItem.taskItem->getText());
	*(taskItem.taskItem) =
		mirai::TodoMdFormat::stringToTask(text.toStdString(), date.toStdString());
	mirai.save();
	todoView.update();
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::updateTodo(int todoIndex, QString state, QString text, QString date)
{
	QMLTaskItem &taskItem = QMLTasks[todoIndex];
	if (state == "DONE") {
		taskItem.taskItem->markAsDone();
	} else if (state == "TODO") {
		taskItem.taskItem->markAsUndone();
	}
	taskItem.taskItem->setText(text.toStdString());
	taskItem.taskItem->setDate(date.toStdString());
	mirai.save();
	todoView.update();
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::removeTodo(int todoIndex)
{
	QMLTaskItem &taskItem = QMLTasks[todoIndex];
	mirai.removeTask(taskItem.taskItem);
	mirai.save();
	todoView.update();
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::hideCompletedTasks(bool shouldHide)
{
	shouldHideCompletedTasks_ = shouldHide;
	rebuildQMLTasksList();
	emit tasksChanged();
}

void Backend::rebuildQMLTasksList()
{
	// TODO MOVE TO ANOTHER FILE
	QMLResources.clear();
	for (auto &resource : mirai.getResources()) {
		QMLResources.push_back({.tasksFile = resource.get()});
	}
	// ----
	QMLTasks.clear();
	std::string lastDate = "";
	std::time_t t = std::time(nullptr);
	std::tm tm = *std::localtime(&t);
	std::stringstream currentDate;
	currentDate << std::put_time(&tm, "%Y-%m-%d");
	for (int i = 0; i < todoView.count(); ++i) {
		mirai::TaskItem &task = todoView[i];
		if (shouldHideCompletedTasks_ && task.getState() == mirai::DONE && task.hasDate() &&
			task.getDate() < currentDate.str()) {
			continue;
		}
		bool shouldShowDate = false;
		if (lastDate != task.getDate()) {
			lastDate = task.getDate();
			shouldShowDate = true;
		}
		QList<QString> qStringTags;
		for (auto &tag : task.getTags()) {
			qStringTags.push_back(QString::fromStdString(tag));
		}
		QMLTasks.push_back(
			{.taskItem = &task, .shouldShowDate = shouldShowDate, .tags = qStringTags}
		);
	}

	QMLTags.clear();
	for (auto &tag : mirai.getTags()) {
		QMLTags.push_back({
			.name = QString::fromStdString(tag),
			.color = getTagColor(QString::fromStdString(tag)),
		});
	}

	QMLActiveTagsFilter.clear();
	for (auto &activeTagFilter : todoView.getActiveTagsFilter()) {
		QMLActiveTagsFilter.push_back(QString::fromStdString(activeTagFilter));
	}

	QMLActiveResourcesFilter.clear();
	for (auto &activeFileFilter : todoView.getActiveFilesFilter()) {
		QMLActiveResourcesFilter.push_back(QString::fromStdString(activeFileFilter));
	}
}

QVariant Backend::getTasks()
{
	return QVariant::fromValue(QMLTasks);
}

QVariant Backend::getTags()
{
	return QVariant::fromValue(QMLTags);
}

QVariant Backend::getActiveTagsFilter()
{
	return QVariant::fromValue(QMLActiveTagsFilter);
}

QVariant Backend::getActiveResourcesFilter()
{
	return QVariant::fromValue(QMLActiveResourcesFilter);
}

bool Backend::shouldHideCompletedTasks()
{
	return shouldHideCompletedTasks_;
}

QVariant Backend::getResources()
{
	return QVariant::fromValue(QMLResources);
}

QString Backend::getTagColor(QString tag)
{
	if (!tagsConfig.contains(tag)) {
		return "#c6d0f5"; // TODO: currently hard coded but should match the default Text color.
	}
	return tagsConfig[tag];
}

void Backend::saveTagsColor(QVariant modifiedTags)
{
	auto aa = modifiedTags.toList();
	for (auto bb : aa) {
		auto cc = bb.toMap();
		std::cout << cc["name"].toString().toStdString() << ": "
				  << cc["color"].toString().toStdString() << std::endl;
		tagsConfig[cc["name"].toString()] = cc["color"].toString();
	}
	rebuildQMLTasksList();
	emit tagsChanged();
	emit tasksChanged();
	saveConfig();
}

void Backend::saveFilesPath(QVariant modifiedFilesPath)
{
	// TODO: Make something better, rework rebuildQMLTasksList ?
	auto paths = modifiedFilesPath.toStringList();
	QMLTags.clear();
	QMLTasks.clear();
	QMLResources.clear();
	emit tagsChanged();
	emit tasksChanged();
	emit filesChanged();
	mirai.unloadAllResources();
	for (auto path : paths) {
		if (path == "") {
			continue;
		}
		auto fileResource =
			std::make_unique<mirai::AndroidFileResource>(mirai::BaseFileResourceConstructor{
				.name = "[Can't load path]", .path = path.toStdString()
			});
		mirai.loadResource(std::move(fileResource));
	}
	todoView.update();
	rebuildQMLTasksList();
	emit tagsChanged();
	emit tasksChanged();
	emit filesChanged();
	saveConfig();
}

void Backend::saveConfig()
{
	QJsonObject rootJson;

	QJsonArray filesJson;
	for (auto &file : mirai.getResources()) {
		auto fsResource = dynamic_cast<mirai::BaseFileResource *>(file.get());
		if (fsResource) {
			filesJson.append(QString::fromStdString(fsResource->getPath()));
		}
	}
	rootJson["files"] = filesJson;

	QJsonObject tagsJson;
	for (auto &tag : tagsConfig.keys()) {
		QJsonObject tagConfig;
		tagConfig["color"] = getTagColor(tag);
		tagsJson[tag] = tagConfig;
	}
	rootJson["tags"] = tagsJson;

	QFile configFile(getConfigFilePath());
	if (!configFile.open(QIODevice::WriteOnly)) {
		qWarning() << "Cannot save config";
		return;
	}
	configFile.write(QJsonDocument(rootJson).toJson());
	configFile.close();
	qDebug() << "Config file saved";
}

QString Backend::getConfigFilePath() const
{
	QString configFolder = QStandardPaths::writableLocation(QStandardPaths::AppConfigLocation);
	return configFolder + "/config.json";
}
